class ParticleSystemAM extends ParticleSystem {
  ParticleSystemAM(float G, float D) {
    super(G,D);
  }
  //Adams-Moulton Predictor-Corrector integrator
  //uses the four previous time steps (3 more than usual)
  //fits a cubic spline to those points to predict next time step
  //this would work faster if the previous velocities were stored in the Particles themselves
  class AdamsMoulton implements Integrator {
    ParticleSystemAM s;
    HashMap velocity1, velocity2, velocity3, velocity4;
    HashMap force1, force2, force3, force4;

    AdamsMoulton(ParticleSystemAM _s) {
      s = _s;
      velocity1 = new HashMap();
      velocity2 = new HashMap();
      velocity3 = new HashMap();
      velocity4 = new HashMap();
      force1 = new HashMap();
      force2 = new HashMap();
      force3 = new HashMap();
      force4 = new HashMap();
    }

    void step(float deltaT) {
      s.applyForces();
      pushForces();
      ArrayList originalPositions = new ArrayList(s.numberOfParticles());
      //save original positions
      for(int i=0;i<s.numberOfParticles();++i) {
        Particle p = s.getParticle(i);
        //need to create a new vector or else if I modify p.position() this changes
        originalPositions.add(new Vector3D(p.position()));
      }
      for(int i=0;i<s.numberOfParticles();++i) {
        Particle p = s.getParticle(i);
        Vector3D v4 = (Vector3D) velocity4.get(p);
        if(v4 == null) { //use midpoint method
          //pEST = p(t) + deltaT*v(t)
          p.position().add(p.velocity().x()*deltaT,p.velocity().y()*deltaT,p.velocity().z()*deltaT);
          p.velocity().add(p.force().x()*deltaT/p.mass(),p.force().y()*deltaT/p.mass(),p.force().z()*deltaT/p.mass());
        } 
        else { //use adams-moulton
          //pEst = p(t) + deltaT/24.0*(-9*v(t-3)+37*v(t-2)-59*v(t-1)+55*v(t))
          Vector3D v3 = (Vector3D) velocity3.get(p);
          Vector3D v2 = (Vector3D) velocity2.get(p);
          Vector3D v1 = (Vector3D) velocity1.get(p);
          Vector3D f4 = (Vector3D) force4.get(p);
          Vector3D f3 = (Vector3D) force3.get(p);
          Vector3D f2 = (Vector3D) force2.get(p);
          Vector3D f1 = (Vector3D) force1.get(p);
          Vector3D delPosition = new Vector3D();
          delPosition.add(-9.0*v4.x(),-9.0*v4.y(),-9.0*v4.z());
          delPosition.add(37.0*v3.x(),37.0*v3.y(),37.0*v3.z());
          delPosition.add(-59.0*v2.x(),-59.0*v2.y(),-59.0*v2.z());
          delPosition.add(55.0*v1.x(),55.0*v1.y(),55.0*v1.z());
          delPosition.multiplyBy(deltaT/24.0);
          p.position().add(delPosition);

          Vector3D delVelocity= new Vector3D();
          delVelocity.add(-9.0*f4.x(),-9.0*f4.y(),-9.0*f4.z());
          delVelocity.add(37.0*f3.x(),37.0*f3.y(),37.0*f3.z());
          delVelocity.add(-59.0*f2.x(),-59.0*f2.y(),-59.0*f2.z());
          delVelocity.add(55.0*f1.x(),55.0*f1.y(),55.0*f1.z());
          delVelocity.multiplyBy(deltaT/24.0/p.mass());
          p.velocity().add(delVelocity);
        }
        //clear forces now that we are done with them
        p.force().clear();
      }
      //get force at new position and velocity
      s.applyForces();
      for(int i=0;i<s.numberOfParticles();++i) {
        Particle p = s.getParticle(i);
        Vector3D originalP = (Vector3D) originalPositions.get(i);
        Vector3D v3 = (Vector3D) velocity3.get(p);
        if(v3 == null) { //use midpoint method
          //p(t+1) = p(t) + deltaT/2.0*(v(t)+vEst)
          Vector3D v1 = (Vector3D) velocity1.get(p);
          Vector3D f1 = (Vector3D) force1.get(p);
          Vector3D tforce = new Vector3D();
          Vector3D tvel = new Vector3D();
          tforce.add(f1);
          tforce.add(p.force());
          tforce.multiplyBy(0.5);
          tvel.add(v1);
          tvel.add(p.velocity());
          tvel.multiplyBy(0.5);
          p.position().set(originalP);
          p.position().add(tvel.x()*deltaT,tvel.y()*deltaT,tvel.z()*deltaT);
          p.velocity().set(v1);
          p.velocity().add(tforce.x()*deltaT,tforce.y()*deltaT,tforce.z()*deltaT);
        } 
        else { //use adams-moulton
          //p(t+1) = p(t) + deltaT/24.0*(v(t-2)-5.0*v(t-1)+19*v(t)+9*vEst)
          Vector3D v2 = (Vector3D) velocity2.get(p);
          Vector3D v1 = (Vector3D) velocity1.get(p);
          Vector3D f3 = (Vector3D) force3.get(p);
          Vector3D f2 = (Vector3D) force2.get(p);
          Vector3D f1 = (Vector3D) force1.get(p);
          Vector3D delPosition = new Vector3D();
          delPosition.add(v3.x(),v3.y(),v3.z());
          delPosition.add(-5.0*v2.x(),-5.0*v2.y(),-5.0*v2.z());
          delPosition.add(19.0*v1.x(),19.0*v1.y(),19.0*v1.z());
          delPosition.add(9.0*p.velocity().x(),9.0*p.velocity().y(),9.0*p.velocity().z());
          delPosition.multiplyBy(deltaT/24.0);
          p.position().set(originalP);
          p.position().add(delPosition);

          Vector3D delVelocity= new Vector3D();
          delVelocity.add(f3.x(),f3.y(),f3.z());
          delVelocity.add(-5.0*f2.x(),-5.0*f2.y(),-5.0*f2.z());
          delVelocity.add(9.0*f1.x(),19.0*f1.y(),19.0*f1.z());
          delVelocity.add(9.0*p.force().x(),9.0*p.force().y(),9.0*p.force().z());
          delVelocity.multiplyBy(deltaT/24.0/p.mass());
          p.velocity().set(v1);
          p.velocity().add(delVelocity);
        }
      }      
      
    }

    void pushForces() {
      for(int i=0;i<s.numberOfParticles();++i) {
        Particle p = s.getParticle(i);
        Object v,f;
        v = velocity3.get(p);
        f = force3.get(p);
        velocity4.put(p,v);
        force4.put(p,f);
        v = velocity2.get(p);
        f = force2.get(p);
        velocity3.put(p,v);
        force3.put(p,f);
        v = velocity1.get(p);
        f = force1.get(p);
        velocity2.put(p,v);
        force2.put(p,f);
        velocity1.put(p,p.velocity());
        force1.put(p,p.force());      
      }    
    }
  }
}

